import scala.virtualization.lms.common.{BaseExp, Base}
import scala.virtualization.lms.internal.Scheduling
import scala.virtualization.lms.util.OverloadHack

trait Example2Ops extends Base with OverloadHack {
  class Matrix[T] // placeholder

  object Matrix {
    def apply[T:Manifest](x: Rep[Int], y: Rep[Int]) = matrix_new[T](x,y)
  }

  // it is not necessary to have separate methods infix_* and matrix_times, but i prefer to keep the syntactic
  // lifting separate from the implementation routines (this way the syntax can be changed to forward to different
  // implementations without changing anything else)
  def infix_*[T](x: Rep[Matrix[T]], y: Rep[Matrix[T]])(implicit mA: Manifest[T], o: Overloaded1) = matrix_times(x,y)
  def infix_*[T](x: Rep[Matrix[T]], y: Rep[Int])(implicit mT: Manifest[T], o: Overloaded2) = matrix_times_scalar(x,y)
  def println[T:Manifest](x: Rep[Matrix[T]]) = matrix_print(x)

  def matrix_new[T:Manifest](x: Rep[Int], y: Rep[Int]): Rep[Matrix[T]]
  def matrix_times[T:Manifest](x: Rep[Matrix[T]], y: Rep[Matrix[T]]): Rep[Matrix[T]]
  def matrix_times_scalar[T:Manifest](x: Rep[Matrix[T]], y: Rep[Int]): Rep[Matrix[T]]
  def matrix_print[T:Manifest](x: Rep[Matrix[T]]): Rep[Unit]
}

trait Example2OpsExp extends Example2Ops with BaseExp {
  case class MatrixNew[T:Manifest](x: Exp[Int], y: Exp[Int]) extends Def[Matrix[T]]
  case class MatrixTimes[T:Manifest](x: Exp[Matrix[T]], y: Exp[Matrix[T]]) extends Def[Matrix[T]]
  case class MatrixTimesScalar[T:Manifest](x: Exp[Matrix[T]], y: Exp[Int]) extends Def[Matrix[T]]

  def matrix_new[T:Manifest](x: Exp[Int], y: Exp[Int]) = MatrixNew[T](x,y)
  def matrix_times[T:Manifest](x: Exp[Matrix[T]], y: Exp[Matrix[T]]) = MatrixTimes(x,y)
  def matrix_times_scalar[T:Manifest](x: Exp[Matrix[T]], y: Exp[Int]) = MatrixTimesScalar(x,y)

  // toggle these to print a string representation of the IR in different ways
  //def matrix_print[T:Manifest](x: Exp[Matrix[T]]) = Predef.println(unravelInline(x))
  def matrix_print[T:Manifest](x: Exp[Matrix[T]]) = unravelMultiline(x)

  private def unravelInline(e: Any): String = e match {
    case s@Sym(id) => findDefinition(s).map(t => t.rhs.getClass.getSimpleName + "(" +
                                                 t.rhs.asInstanceOf[Product].productIterator.toList.map(unravelInline(_)).mkString(",") + ")"
                                           ).getOrElse("")
    case _ => e.toString
  }

  private def unravelMultiline(e: Any): Unit = e match {
    case s@Sym(id) => val tp = findDefinition(s)
                      tp.foreach(t => Predef.println(t.sym.toString + " = " + t.rhs.toString ))
                      tp.foreach(t => t.rhs.asInstanceOf[Product].productIterator.foreach(unravelMultiline(_)))
    case _ =>
  }

}